import os

import pytest
import sqlalchemy.dialects
import sqlalchemy.exc
from jwt import InvalidSignatureError
from sqlalchemy.orm import Session

import rasax.community.database.utils as db_utils
import rasax.community.sql_migrations as migrations
from rasax.community import config, constants
from rasax.community import utils
from rasax.community.database import Role
from tests.unit import conftest
from tests.unit.conftest import insert_test_user, TEST_USER, generate_key_pair


# Overwrite fixture to run per function
@pytest.fixture()
def session():
    session_maker = conftest.session_maker_for_tests()
    session = session_maker()

    migrations.run_migrations(session)

    # Insert general test user which can be used in later tests
    insert_test_user(session, TEST_USER)

    return session_maker()


def test_query_result_to_dict():
    query_result = ["zero", "one", "two", "three"]
    fields_query = [("d", True), ("a.c", True), ("a.b", True), ("b.c.d", True)]

    actual = utils.query_result_to_dict(query_result, fields_query)

    assert actual["d"] == query_result[0]
    assert actual["a"]["c"] == query_result[1]
    assert actual["a"]["b"] == query_result[2]
    assert actual["b"]["c"]["d"] == query_result[3]


def test_foreign_keys_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.enforce_sqlite_foreign_key_constraints(raw_connection, True)
    with pytest.raises(sqlalchemy.exc.IntegrityError):
        db_connection.execute("DROP table {};".format(Role.__tablename__))


def test_turn_off_foreign_keys_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.enforce_sqlite_foreign_key_constraints(raw_connection, False)
    if isinstance(db_connection.dialect, sqlalchemy.dialects.sqlite.dialect):
        db_connection.execute("DROP table {};".format(Role.__tablename__))


def test_turn_off_then_on_foreign_key_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.enforce_sqlite_foreign_key_constraints(raw_connection, False)
    db_utils.enforce_sqlite_foreign_key_constraints(raw_connection, True)

    with pytest.raises(sqlalchemy.exc.IntegrityError):
        db_connection.execute("DROP table {};".format(Role.__tablename__))


@pytest.mark.parametrize(
    "before_truncation, expected",
    [(1.0, 1.0), (1.1234, 1.12), (19231231.00, 19231231.00)],
)
def test_truncate_float(before_truncation: float, expected: float):
    assert utils.truncate_float(before_truncation, decimal_places=2) == expected


def test_verify_bearer_token(tmpdir_factory):
    # get valid token
    private_key_path, public_key_path = generate_key_pair(tmpdir_factory, 0)
    private_key = utils.file_as_bytes(private_key_path)
    public_key = utils.file_as_bytes(public_key_path)

    payload = {"test": "test"}
    encoded = utils.bearer_token(payload, private_key=private_key)

    assert utils.verify_bearer_token(encoded, public_key=public_key) == payload


def test_invalid_bearer_tokens():
    with pytest.raises(ValueError):
        utils.bearer_token({}, private_key="wrong secret")

    with pytest.raises(ValueError):
        utils.encode_jwt({}, private_key="some random key")

    with pytest.raises(ValueError):
        utils.verify_bearer_token("", public_key="wrong key")


def test_jwt_keys_from_file(tmpdir_factory):
    # generate rsa keypair
    private_key_path, public_key_path = generate_key_pair(tmpdir_factory, 0)

    # verification should work
    utils.verify_and_set_jwt_keys_from_file(private_key_path, public_key_path)

    # check that environment variables are set
    assert config.jwt_private_key == utils.file_as_bytes(private_key_path)
    assert config.jwt_public_key == utils.file_as_bytes(public_key_path)

    # wrong key path
    with pytest.raises(SystemExit):
        utils.verify_and_set_jwt_keys_from_file(private_key_path, "some/wrong/path")

    # remove -----BEGIN RSA PRIVATE KEY----- header from private key
    tag = "-----BEGIN RSA PRIVATE KEY-----"
    assert tag.encode() in utils.file_as_bytes(private_key_path)
    os.system("sed -i.bak '/{}/d' {}".format(tag, private_key_path))
    assert tag.encode() not in utils.file_as_bytes(private_key_path)

    private_key = utils.file_as_bytes(private_key_path)
    public_key = utils.file_as_bytes(public_key_path)
    with pytest.raises(ValueError):
        utils.verify_keys(private_key, public_key)

    # generate new private and public keys which don't match each other
    new_private_key_path, _ = generate_key_pair(tmpdir_factory, 1)
    _, new_public_key_path = generate_key_pair(tmpdir_factory, 2)

    new_public_key = utils.file_as_bytes(new_public_key_path)
    new_private_key = utils.file_as_bytes(new_private_key_path)
    with pytest.raises(InvalidSignatureError):
        utils.verify_keys(new_private_key, new_public_key)


def test_generate_key_pair():
    private_key, public_key = utils.generate_rsa_key_pair(2048)
    utils.verify_keys(private_key, public_key)


def test_generate_and_set_jwt_keys():
    utils.generate_and_set_jwt_keys()
    assert config.jwt_public_key is not None
    assert config.jwt_private_key is not None


def test_add_jwt_key_to_dictionary():
    utils.generate_and_set_jwt_keys()
    d = {}
    utils.add_jwt_key_to_dictionary(d)
    assert d["keys"]

    jwt = d["keys"][0]
    assert all(k in jwt for k in ["alg", "key"])

    assert jwt["key"] == config.jwt_public_key
    assert jwt["alg"] == constants.JWT_METHOD
